﻿using System;


namespace WindowsFormsApplication1
{
    class Jacobi
    {
        int maxOrder = 4;
        public double[,] a;
        public double[,] r;
        public double[,] v;

        public Jacobi(int order)
        {
            int i, k;
            maxOrder = order;
            a = new double[order, order];
            r = new double[order, order];
            v = new double[order, order];
            for (i = 0; i < maxOrder; i++)
            {
                for (k = 0; k < maxOrder; k++)
                {
                   
                    a[k, i] = 0.0;
                    v[k, i] = 0.0;
                }
                v[i, i] = 1.0;
            }
        }

        public double[,] MultMatrixRight(double[,] a, double[,] m, int p, int q, int size)
        {
            int i, k;
            double[,] al = new double[size, 2];
            for (i = 0; i < size; i++)
            {
                al[i, 0] = 0.0;
                al[i, 1] = 0.0;
                for (k = 0; k < size; k++)
                {
                    al[i, 0] = al[i, 0] + (a[i, k] * m[k, p]);
                    al[i, 1] = al[i, 1] + (a[i, k] * m[k, q]);
                }
            }
            for (i = 0; i < size; i++)
            {
                a[i, p] = al[i, 0];
                a[i, q] = al[i, 1];
            }
            return a;
        }


        public double[,] MultTranspMatrixLeft(double[,] a, double[,] m, int p, int q, int size)
        {
            int j, k;
            double[,] al = new double[2, size];
            for (j = 0; j < size; j++)
            {
                al[0, j] = 0.0;
                al[1, j] = 0.0;
                for (k = 0; k < size; k++)
                {
                    al[0, j] = al[0, j] + (m[k, p] * a[k, j]);
                    al[1, j] = al[1, j] + (m[k, q] * a[k, j]);
                }
            }
            for (j = 0; j < size; j++)
            {
                a[p, j] = al[0, j];
                a[q, j] = al[1, j];
            }
            return a;
        }


        public void CalcRMatrix(double[,] a, int p, int q, int size)
        {
            int i, j;
            double phi;
            double t;
            phi = (a[q, q] - a[p, p]) / 2 / a[q, p];
            if (phi != 0)
            {
                if (phi > 0)
                    t = 1 / (phi + Math.Sqrt(phi * phi + 1.0));
                else
                    t = 1 / (phi - Math.Sqrt(phi * phi + 1.0));
            }
            else
                t = 1.0;
            for (i = 0; i < size; i++)
            {
                for (j = 0; j < size; j++)
                {
                    r[i, j] = 0.0;
                }
                r[i, i] = 1.0;
            }
            r[p, p] = 1 / Math.Sqrt(t * t + 1.0);
            r[p, q] = t * r[p, p];
            r[q, q] = r[p, p];
            r[q, p] = -r[p, q];
        }


        public void RunJacobi()
        {
            int i, k, count = 0;
            double delta = 10;
            for (i = 0; i < maxOrder; i++)
            {
                for (k = 0; k < maxOrder; k++)
                {

                    r[k, i] = 0.0;
                    v[k, i] = 0.0;
                }
                v[i, i] = 1.0;
            }

            while ((delta > 1.0E-50) && (count < 50))
            {
                delta = 0;
                count++;
                for (i = 0; i < maxOrder; i++)
                {
                    for (k = i + 1; k < maxOrder; k++)
                    {
                        if (a[k, i] != 0.0)
                        {
                            CalcRMatrix(a, i, k, maxOrder);
                            a = MultTranspMatrixLeft(a, r, i, k, maxOrder);
                            a = MultMatrixRight(a, r, i, k, maxOrder);
                            v = MultMatrixRight(v, r, i, k, maxOrder);
                        }
                    }
                    if (i < maxOrder - 1)
                        delta = delta + Math.Abs(a[i + 1, i]);
                }
            }
        }
    }
}
