﻿using System;
using System.Linq;
using System.Collections.ObjectModel;
using System.Windows.Forms;
using System.Diagnostics;

/*Principal components analysis in Gradient descent*/
/*               www.mosismath.com                 */

namespace WindowsFormsApplication1
{
    

    public partial class Mainwin : Form
    {
        int samples = 60;
        const int iterations = 5000;
        const int features = 4;    // number of features (or dimensions) of the input data
        const int redfeatures = 2; // number of features (or dimensions) of the reduced data
        double learning_rate = 0.02;
        double cost = 1000;
        double[] rise = new double[features];
        double[] rise_ori = new double[features];
        double[,] convMatrix = new double[features, redfeatures];
        double offset = 0;
        double offset_ori = 0;
        Data data = new Data();
        Jacobi jacobi = new Jacobi(features);

        //functions for Gradient descend
        double cost_function(Collection<double[]> values, double[] rise)
        {
            int i;
            double cost = 0;
            double[] predictions = update_values(values, rise);
            double[] x;
            double y;

            cost = 0;
            for (i = 0; i < samples; i++)
            {
                x = values[i];
                y = x[redfeatures];
                cost += (y - predictions[i]) * (y - predictions[i]);
            }
            
            cost = cost / samples / 2.0;
            return cost;
        }


        double[] update_values(Collection<double[]> values, double[] rise)
        {
            double[] predictions = new double[samples];
            double[] x;
            for (int i = 0; i < samples; i++)
            {
                predictions[i] = 0;
                x = values[i];
                for (int j = 0; j < redfeatures; j++)
                {
                    predictions[i] = predictions[i] + x[j] * rise[j];
                }
                predictions[i] = predictions[i] + offset;
            }
            return predictions;
        }


        double[] update_rise(Collection<double[]> values, double[] rise, double learning_rate)
        {
            int i, j;
            double[] predictions = update_values(values, rise);
            double[] d_rise = new double[redfeatures];
            double[] x;
            double y;
            double tempSum = 0;

            for (i = 0; i < redfeatures; i++)
            {
                d_rise[i] = 0;
                tempSum = 0;
                for (j = 0; j < samples; j++)
                {
                    x = values.ElementAt(j);
                    y = x[redfeatures];
                    tempSum = tempSum + x[i] * (y - predictions[j]);
                    if(Math.Abs(tempSum) > 1e6)
                    {
                        d_rise[i] = d_rise[i] - tempSum / samples;
                        tempSum = 0;
                    }
                }
                d_rise[i] = d_rise[i] - tempSum / samples;
            }

            for (i = 0; i < redfeatures; i++)
            {
                rise[i] = rise[i] - learning_rate * d_rise[i];
            }
            return rise;
        }


        double[] find_min(Collection<double[]> values, double[] rise, double learning_rate, int maxIers)
        {
            int count = 0;
            double oldCost = -1;
            double divCost = oldCost - cost;
            double[] ret = new double[2];

            while (Math.Abs(divCost) > 1e-6 && count < maxIers)
            {
                rise = update_rise(values, rise, learning_rate);
                cost = cost_function(values, rise);
                divCost = oldCost - cost;
                oldCost = cost;
                count++;
                if (divCost < -0.1)
                {
                    learning_rate = -learning_rate / 2;
                }
            }

            return rise;
        }


        double mean_function(Collection<double[]> values, double[] rise)
        {
            int i;
            double mean = 0;
            double[] predictions = update_values(values, rise);
            double y;

            for (i = 0; i < samples; i++)
            {
                y = values.ElementAt(i)[redfeatures];
                mean += (y - predictions[i]);
            }
            mean = mean / samples / 2;
            return mean;
        }


        // functions for pca
        public double CalcXTX(Collection<double[]> values, int index1, int index2)
        {
            int i;
            double[] x;
            double res = 0;
            double mean1 = 0;
            double mean2 = 0;
            for (i = 0; i < samples; i++)
            {
                x = values.ElementAt(i);
                mean1 = mean1 + x[index1];
                mean2 = mean2 + x[index2];
            }
            mean1 = mean1 / samples;
            mean2 = mean2 / samples;
            for (i = 0; i < samples; i++)
            {
                x = values.ElementAt(i);
                res = res + x[index1] *  x[index2];
            }
            return res/(samples-1);
        }


        void calc_rotation_matrix(int nbOfMins)
        {
            int i, j, k;
            double actMinVal = -1E20;
            double tempMinVal = 1E20;
            int indexOfMin = -1;
            int actIndex = 0;
            for (k = 0; k < nbOfMins; k++)
            {
                tempMinVal = 1E20;
                for (i = 0; i < features; i++)
                {
                    if ((Math.Abs(jacobi.a[i, i]) < tempMinVal) &&(Math.Abs(jacobi.a[i, i]) > actMinVal))
                    {
                        tempMinVal = Math.Abs(jacobi.a[i, i]);
                        indexOfMin = i;
                    }
                }
                actMinVal = tempMinVal;
            }

            for (i = 0; i < features; i++)
            {
                if (Math.Abs(jacobi.a[i, i]) > actMinVal)
                {
                    for (j = 0; j < features; j++)
                    {
                        convMatrix[j, actIndex] = jacobi.v[j, i];
                    }
                    actIndex++;
                }
            }
        }


        double Calc_Y_Reduced(Collection<double[]> values, int index)
        {
            int i;
            double[] x;
            double[] xRed = new double[2];
            x = values.ElementAt(index);
            for (i = 0; i < x.Length - 1; i++)
            {
                xRed[0] = xRed[0] + x[i] * convMatrix[i, 0];
                xRed[1] = xRed[1] + x[i] * convMatrix[i, 1];
            }
            return xRed[0] * rise[0] + xRed[1] * rise[1] + offset;
        }


        void Calc_invers_rotation()
        {
            int i, j;
            offset_ori = offset;
            for (i = 0; i < features; i++)
            {
                rise_ori[i] = 0;
                for (j = 0; j < redfeatures; j++)
                {
                    rise_ori[i] = rise_ori[i] + rise[j] * convMatrix[i, j];
                    offset_ori = offset_ori - (data.mean[i] / data.deviation[i] * rise[j] * convMatrix[i, j]);
                }
                rise_ori[i] = rise_ori[i] / data.deviation[i];
            }
        }


        // main window functions
        public Mainwin()
        {
            int i, j;
            DataGridViewCell cell;

            InitializeComponent();
            data.readData(Application.StartupPath + "\\data.csv");
            gResult.ColumnCount = features + 1;
            gResult.RowCount = 2;
            gResult.Rows[0].HeaderCell.Value = "red";
            gResult.Rows[1].HeaderCell.Value = "ori";
            samples = data.values.Count;

            dgData.RowCount = samples;
            dgData.ColumnCount = features + 3;
            for (j = 0; j < features; j++)
                dgData.Columns[j].HeaderText = "x" + (j+1).ToString();
            dgData.Columns[j].HeaderText = "y";
            dgData.Columns[j+1].HeaderText = "f(x)";
            dgData.Columns[j + 2].HeaderText = "delta [%]";
            for (i = 0; i < samples; i++)
            {
                dgData.Rows[i].HeaderCell.Value = (i+1).ToString();
    
                for (j = 0; j <= features; j++)
                {
                    cell = dgData[j, i];
                    cell.Value = Math.Round(data.values.ElementAt(i)[j], 4).ToString();
                }
            }
        }

        private void bt_Calc_Click(object sender, EventArgs e)
        {
            int i, j, k;
            int count = 0;
            double oldCost = 1e60;
            double divCost = 10;
            double lastCost = 10;
            double step = 0.5;
            double y;
            DataGridViewCell cell;
            Collection<double[]> reducedValues = new Collection<double[]>();

            data.calcMean(features);
            data.normalize(features);
            
            // build the rotation matrix
            for (i = 0; i < features; i++)
            {
                for (j = 0; j < features; j++)
                {
                    jacobi.a[i, j] = CalcXTX(data.values, i, j);
                }
            }
            jacobi.RunJacobi();
            calc_rotation_matrix(features - redfeatures);
           
            // build the reduced data    
            double[] tempVals;
            for (k = 0; k < samples; k++)
            {
                double[] reducedLine = new double[redfeatures + 1];
                tempVals = data.values.ElementAt(k);
                for (i = 0; i < redfeatures; i++)
                {
                    reducedLine[i] = 0;
                    for (j = 0; j < features; j++)
                        reducedLine[i] = reducedLine[i] + tempVals[j] * convMatrix[j, i];
                }
                reducedLine[redfeatures] = tempVals[features];
                reducedValues.Add(reducedLine);
            }

            // compute gradient descend
            offset = 0;
            while ((Math.Abs(divCost) > 0.000001) && (count < 1000))
            {
                rise = find_min(reducedValues, rise, learning_rate, iterations);

                divCost = oldCost - cost;
                if (divCost < -0.001)
                {
                    step = -step / 2.0;
                }
                lastCost = oldCost;
                oldCost = cost;
                offset = offset + mean_function(reducedValues, rise); 
                count++;
            }

            Calc_invers_rotation();
            // Display the computed data
            for (i = 0; i < features; i++)
            {
                gResult.Columns[i].HeaderText = "r" + (i + 1).ToString();
                cell = gResult[i, 0];
                cell.Value = Math.Round(rise[i], 4).ToString();
                cell = gResult[i, 1];
                cell.Value = Math.Round(rise_ori[i], 4).ToString();
            }
            gResult.Columns[i].HeaderText = "offset";
            cell = gResult[i, 0];
            cell.Value = Math.Round(offset, 4).ToString();
            cell = gResult[i, 1];
            cell.Value = Math.Round(offset_ori, 4).ToString();

            tbCost.Text = Math.Round(cost, 8).ToString();
          
            for (i = 0; i < samples; i++)
            {
                y = Calc_Y_Reduced(data.values, i);
                cell = dgData[features + 1, i];
                cell.Value = Math.Round(y, 4).ToString();
                 if (reducedValues.ElementAt(i)[redfeatures] != 0)
                 {
                     cell = dgData[features + 2, i];
                     cell.Value = Math.Round((y - reducedValues.ElementAt(i)[redfeatures]) * (y - reducedValues.ElementAt(i)[redfeatures]) / reducedValues.ElementAt(i)[redfeatures] * 100, 4).ToString();

                 }
            }
        }
    }
}
